﻿using Janus.Core.Importers;
using Janus.Core.Models;
using Janus.Core.Models.Results;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Janus.Analyzers.Counter
{
    public class CounterAnalyzer
    {
        /// <summary>
        /// Set of unique packages
        /// </summary>
        HashSet<DependencyItem> packagesUniqueSet = new HashSet<DependencyItem>();

        /// <summary>
        /// Result of the last run of analysis
        /// </summary>
        public AnalysisResultModel AnalyzerResult { get; set; }

        ////BFS fields

        Queue<DependencyItem> bfsToBeAnalysed = new Queue<DependencyItem>();

        //Key:   Visited dependency refference
        HashSet<DependencyItem> bfsVisitedDependencies = new HashSet<DependencyItem>();


        public CounterAnalyzer()
        {

        }

        /// <summary>
        /// Processes the second part of deserialization and redirects the analysis to the data specific methods
        /// </summary>
        /// <param name="dataWrapper">Data wrapper to be redirected</param>
        public void ProcessDataWrapper(DataSerializationWrapper dataWrapper)
        {
            if (dataWrapper.DataType == null)
            {
                //NULL!!!!
                Console.Error.WriteLine($"ERROR: Datatype in the wrapper is null, cannot deserialize!");
                return;
            }
            else if (dataWrapper.DataType == new List<DependencyItem>().GetType().ToString())
            {
                ImportResult<IList<DependencyItem>> importedData = JsonImporter.Instance.ImportData<IList<DependencyItem>>(dataWrapper);
                if (importedData.OperationSuccessful)
                {
                    dataWrapper.Data = importedData.Result;
                    AnalyzeDependencies(importedData.Result);
                }
                else
                {
                    //Import problem .... end execution
                    foreach (var err in importedData.Errors)
                    {
                        Console.Error.WriteLine(err);
                    }
                    return;
                }
            }
            else if (dataWrapper.DataType == new List<ProjectItem>().GetType().ToString())
            {
                ImportResult<IList<ProjectItem>> importedData = JsonImporter.Instance.ImportData<IList<ProjectItem>>(dataWrapper);
                if (importedData.OperationSuccessful)
                {
                    dataWrapper.Data = importedData.Result;
                    AnalyzeProjects(importedData.Result);
                }
                else
                {
                    //Import problem .... end execution
                    foreach (var err in importedData.Errors)
                    {
                        Console.Error.WriteLine(err);
                    }
                    return;
                }
            }
            else if (dataWrapper.DataType == new List<SolutionItem>().GetType().ToString())
            {
                ImportResult<IList<SolutionItem>> importedData = JsonImporter.Instance.ImportData<IList<SolutionItem>>(dataWrapper);
                if (importedData.OperationSuccessful)
                {
                    dataWrapper.Data = importedData.Result;
                    AnalyzeSolutions(importedData.Result);
                }
                else
                {
                    //Import problem .... end execution
                    foreach (var err in importedData.Errors)
                    {
                        Console.Error.WriteLine(err);
                    }
                    return;
                }
            }
            else
            {
                //Unknown ...
                Console.Error.WriteLine($"ERROR: Datatype in the wrapper is unknown, cannot deserialize!");
                return;
            }

            //Should be imported fully now ... start the analysis
            StartDependencyAnalysis();
        }


        /// <summary>
        /// Performs the analysis queue for each solution of the given list of the solutions
        /// Internally calls <see cref="AnalyzeSolution"/>
        /// </summary>
        /// <param name="solutions">Solutions to analyze</param>
        private void AnalyzeSolutions(IList<SolutionItem> solutions)
        {
            foreach (var solution in solutions)
            {
                AnalyzeSolution(solution);
            }
        }

        /// <summary>
        /// Performs the analysis queue for each project in the given solution
        /// Internally calls <see cref="AnalyzeProject"/>
        /// </summary>
        /// <param name="solution">Solution to analyze</param>
        private void AnalyzeSolution(SolutionItem solution)
        {
            foreach (var projectEdge in solution.Projects)
            {
                if (projectEdge.Link != null)
                {
                    AnalyzeProject(projectEdge.Link);
                }
            }
        }

        /// <summary>
        /// Performs the analysis queue for each project of the given list of the projects
        /// Internally calls <see cref="AnalyzeProject"/>
        /// </summary>
        /// <param name="projects">Projects to analyze</param>
        private void AnalyzeProjects(IList<ProjectItem> projects)
        {
            foreach (var project in projects)
            {
                AnalyzeProject(project);
            }
        }

        /// <summary>
        /// Enques the direct dependencies of the project to the work queue
        /// </summary>
        /// <param name="project">Projects to analyze</param>
        private void AnalyzeProject(ProjectItem project)
        {
            foreach (var dependencyEdge in project.Dependencies)
            {
                if (dependencyEdge.Link != null)
                {
                    bfsToBeAnalysed.Enqueue(dependencyEdge.Link); //Add to the work list
                }
            }
        }

        /// <summary>
        /// Enques each dependency to the work queue
        /// </summary>
        /// <param name="dependencies">Dependencies to analyze</param>
        private void AnalyzeDependencies(IList<DependencyItem> dependencies)
        {
            foreach (var dependency in dependencies)
            {
                bfsToBeAnalysed.Enqueue(dependency); //Add to the work list
                //AnalyzeDependency(dependency); 
            }
        }

        /// <summary>
        /// Performs the dependency analysis with the filled work queue of dependencies 
        /// </summary>
        private void StartDependencyAnalysis()
        {
            AnalyzerResult = new AnalysisResultModel();

            while (bfsToBeAnalysed.Any())
            {
                DependencyItem current = bfsToBeAnalysed.Dequeue();

                if (bfsVisitedDependencies.Contains(current))
                {
                    //Already visited!!!
                    //Console.WriteLine("Been there, done that!");
                    continue;
                }

                bfsVisitedDependencies.Add(current); //Mark as visited

                //Process
                AnalyzeDependency(current);

                //Add subdependencies to the work queue
                foreach (var subDependencyEdge in current.Dependencies)
                {
                    DependencyItem subDependency = subDependencyEdge.Link;
                    if (subDependency == null)
                    {
                        //SKIP null ...
                        continue;
                    }
                    if (bfsVisitedDependencies.Contains(subDependency))
                    {
                        //Already visited ...
                    }
                    else
                    {
                        //This is quite slow ... iterating the queue ... 
                        //Better approach could be importing them to the queue, but during the visisted check they wont be processed anyway ... 
                        //Checking against hash is much faster!!!!
                        /*//New dependency, has to be processed .. or already in the queue?
                        if (bfsToBeAnalysed.Contains(subDependency))
                        {
                            //Is included in to do list ... just reference
                        }
                        else
                        {
                            //Completely new ... will be included in queue
                            bfsToBeAnalysed.Enqueue(subDependency);        //Add to the to process queue

                        }*/
                        bfsToBeAnalysed.Enqueue(subDependency);        //Add to the to process queue
                    }
                }
            }

            //Finalize the obtained results
            FinalizeAnalysis();
        }

        /// <summary>
        /// Analyzes one unique dependency of the given data
        /// </summary>
        /// <param name="dependency">Dependency to be analyzed</param>
        private void AnalyzeDependency(DependencyItem dependency)
        {
            //This is the core of the analysis
            //Here will be executed all the analysis 

            packagesUniqueSet.Add(dependency);
        }

        /// <summary>
        /// Annotates the given data with the version conflict warning mark
        /// Adds to the text the versions of the other conflicting packages
        /// </summary>
        public void FinalizeAnalysis()
        {
            foreach (var package in packagesUniqueSet)
            {
                if (package.AnnotationInfo.Count > 0)
                {
                    AnalyzerResult.AnnotatedUniquePackagesCount++;
                }

                AnalyzerResult.UniquePackagesCount++;
            }
        }
    }
}
