﻿using GiGraph.Dot.Entities.Edges;
using GiGraph.Dot.Entities.Graphs;
using GiGraph.Dot.Entities.Nodes;
using GiGraph.Dot.Extensions;
using GiGraph.Dot.Types.Nodes;
using Janus.Core.Interfaces;
using Janus.Core.Models;
using Janus.Core.Models.Results;
using Janus.Core.Utils;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;

namespace Janus.Core.Exporters
{
    /// <summary>
    /// This class is used for data export to .dot graph format.
    /// </summary>
    public class DotExporter : IExporter
    {
        /// <summary>
        /// Private instance of the DotExporter used as singleton
        /// </summary>
        private static DotExporter _instance = null;

        /// <summary>
        /// Private constructor, to hide from outer code
        /// </summary>
        private DotExporter()
        {
        }

        /// <summary>
        /// Returns the instance of the singleton
        /// </summary>
        public static DotExporter Instance
        {
            get
            {
                if (_instance == null)
                {
                    _instance = new DotExporter();
                }
                return _instance;
            }
        }

        #region Interface implementation

        /// <summary>
        /// This method performs export of the given object to the DOT format and will be saved to desired location
        /// If the output file is not specified, saving is not performed. Output is stored in the ExportResult only.
        /// </summary>
        /// <typeparam name="T">Type of the object to be exported</typeparam>
        /// <param name="dataToExport">Data object to be exported</param>
        /// <param name="outPath">Path to the output file, null saving is ommited</param>
        /// <returns>Result of the operation</returns>
        public ExportResult ExportData<T>(T dataToExport, string outPath = null)
        {
            ExportResult result = new ExportResult();
            GenericResult<DotGraph> graphResult = null;
            DotGraph graph = new DotGraph(directed: true);

            if (dataToExport is List<DependencyItem> dependencies)
            {
                graphResult = ProcessDependenciesGraph(dependencies, graph);    //Creates graph for the list of dependencies only
            }
            else if (dataToExport is List<ProjectItem> projects)
            {
                graphResult = ProcessProjectsGraph(projects, graph);            //Creates graph for the list of project, including all of its dependencies
            }
            else if (dataToExport is List<SolutionItem> solutions)
            {
                graphResult = ProcessSolutionsGraph(solutions, graph);          //Creates graph for the list of solutions, all of the projects in those solutions, including all of its dependencies
            }
            else
            {
                result.OperationSuccessful = false;
                result.Errors.Add($"Unknown datatype for dot conversion! (Datatype: {dataToExport.GetType()})");
                return result;
            }

            //Validate if result is OK

            if (graphResult == null)
            {
                result.OperationSuccessful = false;
                result.Errors.Add($"Conversion to dot format returned no result! (Datatype: {dataToExport.GetType()})");
                return result;
            }
            if (!graphResult.OperationSuccessful)
            {
                result.OperationSuccessful = false;
                result.Errors = graphResult.Errors; //Forward errors
                return result;
            }
            if (graphResult.Result == null)
            {
                result.OperationSuccessful = false;
                result.Errors.Add($"Dot graph export object is null! (Datatype: {dataToExport.GetType()})");
                return result;
            }

            result.ExportData = graphResult.Result.Build(); //Builds the graph and results the string

            //Operation OK, Proceed to saving
            if (outPath != null)
            {
                try
                {
                    File.WriteAllText(outPath, result.ExportData);
                    result.ExportPath = outPath;
                    result.ExportedToFile = true;
                }
                catch (Exception e)
                {
                    result.OperationSuccessful = false;
                    result.Errors.Add(e.Message);
                    return result;
                }
            }
            else
            {
                result.ExportedToFile = false;
            }

            result.OperationSuccessful = true;

            return result;
        }

        #endregion

        #region Private methods

        /// <summary>
        /// Performs the graph creation for the list of dependencies
        /// Has to obtain all the unique items of the graph
        /// </summary>
        /// <param name="dependencies">List of dependencies to be converted to the graph</param>
        /// <param name="graph">Graph instace to be used for graph</param>
        /// <returns>Operation result</returns>
        private GenericResult<DotGraph> ProcessDependenciesGraph(List<DependencyItem> dependencies, DotGraph graph)
        {
            GenericResult<DotGraph> result = new GenericResult<DotGraph>();

            Tuple<List<DependencyItem>, List<DependencyEdge>> uniqueItems = ProcessDependencies(dependencies, true);

            AddDependenciesToGraph(graph, uniqueItems.Item1, uniqueItems.Item2);

            result.OperationSuccessful = true;
            result.Result = graph;

            return result;
        }

        /// <summary>
        /// Creates list of the unique dependencies and edges for the list of dependencies
        /// </summary>
        /// <param name="dependencies">List of dependencies to be analyzed</param>
        /// <param name="includeRoot">Optional, if the root is also required</param>
        /// <returns>Tuple, first item is the list of unique dependencies, second item is the list of the uniquie edges</returns>
        private Tuple<List<DependencyItem>, List<DependencyEdge>> ProcessDependencies(List<DependencyItem> dependencies, bool includeRoot = false)
        {
            HashSet<DependencyItem> deps = new HashSet<DependencyItem>();
            HashSet<DependencyEdge> edges = new HashSet<DependencyEdge>();

            foreach (var dependency in dependencies)
            {
                if (includeRoot)
                {
                    deps.Add(dependency);
                }
                Tuple<List<DependencyItem>, List<DependencyEdge>> uniqueItems = dependency.GetUniqueNodesAndEdges();
                //merge results
                uniqueItems.Item1.ForEach(x => deps.Add(x));
                uniqueItems.Item2.ForEach(x => edges.Add(x));
            }

            return Tuple.Create(deps.ToList(), edges.ToList());
        }

        /// <summary>
        /// Adds the dependencies to the graph
        /// First imports the dependencies, as nodes
        /// Then import dependencyEdges as edges
        /// 
        /// As the id is considered the combination of the ID and version atributes 
        /// </summary>
        /// <param name="graph">Graph, to be the data imported into</param>
        /// <param name="deps">List of dependencies to be imported</param>
        /// <param name="edges">List of dependencyEdges to be imported</param>
        private void AddDependenciesToGraph(DotGraph graph, List<DependencyItem> deps, List<DependencyEdge> edges)
        {
            foreach (var dependency in deps)
            {
                string graphNodeLabel = $"{dependency.Id}\n{dependency.Version}";
                string graphToolTip = $"Id:         {dependency.Id}\n" +
                                      $"Version:    {dependency.Version}\n" +
                                      $"License:    {dependency.License}\n" +
                                      $"Summary:    {dependency.Summary}\n" +
                                      $"Link:       {dependency.Link}\n";


                bool errMark = false;
                bool warnMark = false;
                bool infoMark = false;

                if (dependency.AnnotationInfo.Count > 0)
                {
                    graphToolTip += $"Annotations (one per line):     \n";

                    foreach (var annotation in dependency.AnnotationInfo)
                    {
                        if (annotation.InformationType == AnnotationInfoItem.InformationLevel.Err)
                        {
                            errMark = true;
                        }
                        else if (annotation.InformationType == AnnotationInfoItem.InformationLevel.Warn)
                        {
                            warnMark = true;
                        }
                        else if (annotation.InformationType == AnnotationInfoItem.InformationLevel.Info)
                        {
                            infoMark = true;
                        }

                        graphToolTip += $" - {annotation.ToString()}\n";

                    }
                }

                graph.Nodes.Add($"{dependency.Id}{dependency.Version}", node =>
                {
                    node.Label = graphNodeLabel;
                    node.Shape = DotNodeShape.Oval;
                    node.Tooltip = graphToolTip;
                    if (errMark)
                    {
                        node.Color = Color.Red;
                    }
                    else if (warnMark)
                    {
                        node.Color = Color.Yellow;
                    }
                    else if (infoMark)
                    {
                        node.Color = Color.Blue;
                    }
                });
            }

            foreach (var edge in edges)
            {
                string graphEdgeLabel = $"{edge.RequiredVersionsText}";
                string graphToolTip = $"{edge.Parent.Id}{edge.Parent.Version}->{edge.Link.Id}{edge.Link.Version}\n";

                bool errMark = false;
                bool warnMark = false;
                bool infoMark = false;

                if (edge.AnnotationInfo.Count > 0)
                {
                    graphToolTip += $"Annotations (one per line):     \n";

                    foreach (var annotation in edge.AnnotationInfo)
                    {
                        if (annotation.InformationType == AnnotationInfoItem.InformationLevel.Err)
                        {
                            errMark = true;
                        }
                        else if (annotation.InformationType == AnnotationInfoItem.InformationLevel.Warn)
                        {
                            warnMark = true;
                        }
                        else if (annotation.InformationType == AnnotationInfoItem.InformationLevel.Info)
                        {
                            infoMark = true;
                        }

                        graphToolTip += $" - {annotation.ToString()}\n";
                    }
                }

                graph.Edges.Add($"{edge.Parent.Id}{edge.Parent.Version}", $"{edge.Link.Id}{edge.Link.Version}", graphEdge =>
                {
                    graphEdge.Label = graphEdgeLabel;
                    graphEdge.Tooltip = graphToolTip;
                    if (errMark)
                    {
                        graphEdge.Color = Color.Red;
                    }
                    if (warnMark)
                    {
                        graphEdge.Color = Color.Yellow;
                    }
                    else if (infoMark)
                    {
                        graphEdge.Color = Color.Blue;
                    }
                });
            }
        }

        /// <summary>
        /// Performs the graph creation for the list of projects and all its dependencies
        /// Has to obtain all the unique items of the graph
        /// 
        /// First adds project nodes
        /// Then adds dependencies using method <see cref="AddDependenciesToGraph"/>
        /// Lastly adds project to dependency edges
        /// </summary>
        /// <param name="projects">List of projects to be converted to the graph</param>
        /// <param name="graph">Graph instace to be used for graph</param>
        /// <returns>Operation result</returns>
        private GenericResult<DotGraph> ProcessProjectsGraph(List<ProjectItem> projects, DotGraph graph)
        {
            GenericResult<DotGraph> result = new GenericResult<DotGraph>();

            HashSet<DependencyItem> deps = new HashSet<DependencyItem>();
            HashSet<DependencyEdge> edges = new HashSet<DependencyEdge>();
            List<DotEdge> projectToDepsEdges = new List<DotEdge>();


            foreach (var project in projects)
            {
                graph.Nodes.Add($"{project.Name}", node =>
                {
                    node.Label = $"{project.Name}";
                    node.Shape = DotNodeShape.Box;
                });
                foreach (var dependencyEdge in project.Dependencies)
                {
                    if(dependencyEdge.Link != null)
                    {
                        var dependency = dependencyEdge.Link;
                        deps.Add(dependency);
                        projectToDepsEdges.Add(new DotEdge($"{project.Name}", $"{dependency.Id}{dependency.Version}") { 
                            Label = dependencyEdge.RequiredVersionsText
                        });
                        Tuple<List<DependencyItem>, List<DependencyEdge>> uniqueItems = dependency.GetUniqueNodesAndEdges();
                        //merge results
                        uniqueItems.Item1.ForEach(x => deps.Add(x));
                        uniqueItems.Item2.ForEach(x => edges.Add(x));
                    }                 
                }
            }

            AddDependenciesToGraph(graph, deps.ToList(), edges.ToList());

            foreach (var projectToDepEdge in projectToDepsEdges)
            {
                graph.Edges.Add(projectToDepEdge);
            }

            result.OperationSuccessful = true;
            result.Result = graph;

            return result;
        }

        /// <summary>
        /// Performs the graph creation for the list of solutions, along with all of the projects included and all its dependencies
        /// Has to obtain all the unique items of the graph
        /// 
        /// First adds solution nodes
        /// Then adds project nodes
        /// Then adds dependencies using method <see cref="AddDependenciesToGraph"/>
        /// Then adds project to dependency edges
        /// Lastly adds solution to project edges
        /// </summary>
        /// <param name="solutions">List of solutions to be converted to the graph</param>
        /// <param name="graph">Graph instace to be used for graph</param>
        /// <returns>Operation result</returns>
        private GenericResult<DotGraph> ProcessSolutionsGraph(List<SolutionItem> solutions, DotGraph graph)
        {
            GenericResult<DotGraph> result = new GenericResult<DotGraph>();

            HashSet<DependencyItem> deps = new HashSet<DependencyItem>();
            HashSet<DependencyEdge> edges = new HashSet<DependencyEdge>();
            HashSet<ProjectItem> projects = new HashSet<ProjectItem>();
            List<DotEdge> projectToDepsEdges = new List<DotEdge>();
            List<DotEdge> solutionsToProjectsEdges = new List<DotEdge>();



            foreach (var solution in solutions)
            {
                graph.Nodes.Add($"{solution.Name}", node =>
                {
                    node.Label = $"{solution.Name}";
                    node.Shape = DotNodeShape.Hexagon;
                });

                foreach (var projectEdge in solution.Projects)
                {
                    var project = projectEdge.Link;
                    if (project == null)
                    {
                        continue; //Skip problem
                    }
                    solutionsToProjectsEdges.Add(new DotEdge($"{solution.Name}", $"{project.Name}"));
                    if (projects.Contains(project))
                    {
                        continue; //Skip, add only edge
                    }
                    graph.Nodes.Add($"{project.Name}", node =>
                    {
                        node.Label = $"{project.Name}";
                        node.Shape = DotNodeShape.Box;
                    });

                    foreach (var dependencyEdge in project.Dependencies)
                    {
                        if (dependencyEdge.Link != null)
                        {
                            var dependency = dependencyEdge.Link;
                            deps.Add(dependency);
                            projectToDepsEdges.Add(new DotEdge($"{project.Name}", $"{dependency.Id}{dependency.Version}")
                            {
                                Label = dependencyEdge.RequiredVersionsText
                            });
                            Tuple<List<DependencyItem>, List<DependencyEdge>> uniqueItems = dependency.GetUniqueNodesAndEdges();
                            //merge results
                            uniqueItems.Item1.ForEach(x => deps.Add(x));
                            uniqueItems.Item2.ForEach(x => edges.Add(x));
                        }
                    }
                }
            }

            AddDependenciesToGraph(graph, deps.ToList(), edges.ToList());

            //Adds projects edges
            foreach (var projectToDepEdge in projectToDepsEdges)
            {
                graph.Edges.Add(projectToDepEdge);
            }

            //Adds solutions edges
            foreach (var solutionToProjectEdge in solutionsToProjectsEdges)
            {
                graph.Edges.Add(solutionToProjectEdge);
            }

            result.OperationSuccessful = true;
            result.Result = graph;

            return result;
        }

        #endregion
    }
}
