﻿using Janus.Core.Interfaces;
using Janus.Core.Models;
using Janus.Core.Models.Results;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace Janus.Core.Exporters
{
    /// <summary>
    /// This class is used for data export to .json format.
    /// Data are always wrapped using the <see cref="DataSerializationWrapper"/>
    /// </summary>
    public class JsonExporter : IExporter
    {
        /// <summary>
        /// Private instance of the JsonExporter used as singleton
        /// </summary>
        private static JsonExporter _instance = null;

        /// <summary>
        /// Private constructor, to hide from outer code
        /// </summary>
        private JsonExporter()
        {
        }

        /// <summary>
        /// Returns the instance of the singleton
        /// </summary>
        public static JsonExporter Instance
        {
            get
            {
                if (_instance == null)
                {
                    _instance = new JsonExporter();
                }
                return _instance;
            }
        }

        #region Interface implementation

        /// <summary>
        /// This method performs serialization of the given object to json format and will be saved to desired location
        /// If the output file is not specified, saving is not performed. Output is stored in the ExportResult only.
        /// </summary>
        /// <typeparam name="T">Type of the object to be serialized</typeparam>
        /// <param name="dataToExport">Data object to be serialized</param>
        /// <param name="outPath">Path to the output file, null saving is ommited</param>
        /// <returns>Result of the operation</returns>
        public ExportResult ExportData<T>(T dataToExport, string outPath = null)
        {
            ExportResult result = new ExportResult();
            DataSerializationWrapper wrapperData = new DataSerializationWrapper();

            wrapperData.Version = Consts.UniversalDataFormatVersion;
            wrapperData.DataType = dataToExport.GetType().ToString();
            wrapperData.Data = dataToExport;

            try
            {
                result.ExportData = JsonConvert.SerializeObject(wrapperData, Formatting.Indented,
                                          new JsonSerializerSettings { PreserveReferencesHandling = PreserveReferencesHandling.Objects });
            }
            catch (Exception e)
            {
                //Exception during serialization
                result.OperationSuccessful = false;
                result.Errors.Add(e.Message);
                return result;
            }

            // Try to save file, null means no file saving is required
            if (outPath != null)
            {
                try
                {
                    File.WriteAllText(outPath, result.ExportData);
                    result.ExportPath = outPath;
                    result.ExportedToFile = true;
                }
                catch (Exception e)
                {
                    result.OperationSuccessful = false;
                    result.Errors.Add(e.Message);
                    return result;
                }
            }
            else
            {
                result.ExportedToFile = false;
            }

            result.OperationSuccessful = true;

            return result;
        }

        #endregion

        #region Private methods


        #endregion

    }
}
