﻿using Janus.Core.Interfaces;
using Janus.Core.Models.Results;
using Janus.Core.Models;
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace Janus.Core.Exporters
{
    /// <summary>
    /// This class is used for data export to text format (Human friendly)
    /// Data are printed using the pretty print wih basic information available
    /// </summary>
    public class TextExporter : IExporter
    {
        /// <summary>
        /// Private instance of the TextExporter used as singleton
        /// </summary>
        private static TextExporter _instance = null;

        /// <summary>
        /// Private constructor, to hide from outer code
        /// </summary>
        private TextExporter()
        {
        }

        /// <summary>
        /// Returns the instance of the singleton
        /// </summary>
        public static TextExporter Instance
        {
            get
            {
                if (_instance == null)
                {
                    _instance = new TextExporter();
                }
                return _instance;
            }
        }

        #region Interface implementation

        /// <summary>
        /// This method performs pretty print of the given object to text format and will be saved to desired location
        /// If the output file is not specified, saving is not performed. Output is stored in the ExportResult only.
        /// </summary>
        /// <typeparam name="T">Type of the object to be pretty printed</typeparam>
        /// <param name="dataToExport">Data object to be pretty printed</param>
        /// <param name="outPath">Path to the output file, null saving is ommited</param>
        /// <returns>Result of the operation</returns>
        public ExportResult ExportData<T>(T dataToExport, string outPath = null)
        {
            ExportResult result = new ExportResult();

            if (dataToExport is List<DependencyItem> dependencies)
            {
                foreach (var dependency in dependencies)
                {
                    result.ExportData += dependency.PrintPretty("", true);
                }
            }
            else if (dataToExport is List<ProjectItem> projects)
            {
                foreach (var project in projects)
                {
                    result.ExportData += project.PrintPretty("", true);
                }
            }
            else if (dataToExport is List<SolutionItem> solutions)
            {
                foreach (var solution in solutions)
                {
                    result.ExportData += solution.PrintPretty("", true);
                }
            }
            else
            {
                result.OperationSuccessful = false;
                result.Errors.Add($"Unknown datatype for text conversion! (Datatype: {dataToExport.GetType()})");
                return result;
            }

            // Try to save file, null means no file saving is required
            if (outPath != null)
            {
                try
                {
                    File.WriteAllText(outPath, result.ExportData);
                    result.ExportPath = outPath;
                    result.ExportedToFile = true;
                }
                catch (Exception e)
                {
                    result.OperationSuccessful = false;
                    result.Errors.Add(e.Message);
                    return result;
                }
            }
            else
            {
                result.ExportedToFile = false;
            }

            result.OperationSuccessful = true;

            return result;
        }

        #endregion
    }
}
