﻿using Janus.Core.Interfaces;
using System;
using System.Collections.Generic;
using System.Text;

namespace Janus.Core.Models
{
    /// <summary>
    /// This class is container for all the parameters needed for a depende item.
    /// Identification of the unique item is based on dependency type.
    /// For instance: Nuget has unique ID
    /// </summary>
    public class DependencyItem : IComponentItem, IComparable<DependencyItem>
    {

        /// <summary>
        /// Constructs and initializes the instance
        /// </summary>
        public DependencyItem()
        {
            Dependencies = new List<DependencyEdge>();
            AnnotationInfo = new List<AnnotationInfoItem>();
            Authors = new List<string>();
            CustomAttributes = new Dictionary<string, object>();
        }

        /// <summary>
        /// Uniquie ID in the package registry
        /// </summary>
        public string Id { get; set; }

        /// <summary>
        /// Dependency name
        /// </summary>
        public string Name { get; set; }

        /// <summary>
        /// Dependency version, format is same, as for the package registry
        /// </summary>
        public string Version { get; set; }

        /// <summary>
        /// Package registry type
        /// </summary>
        public string RegistryType { get; set; }

        /// <summary>
        /// List of authors of the package
        /// </summary>
        public IList<string> Authors { get; set; }

        /// <summary>
        /// License information
        /// </summary>
        public string License { get; set; }

        /// <summary>
        /// License link
        /// </summary>
        public string LicenseLink { get; set; }

        /// <summary>
        /// Summary text of the package
        /// Should be one line description
        /// </summary>
        public string Summary { get; set; }

        /// <summary>
        /// Textual description of the package
        /// </summary>
        public string Description { get; set; }

        /// <summary>
        /// Package link
        /// </summary>
        public string Link { get; set; }

        /// <summary>
        /// Package download link
        /// </summary>
        public string PackageDownloadLink { get; set; }

        /// <summary>
        /// Link to the repository of the source code
        /// </summary>
        public string SourceCodeLink { get; set; }

        /// <summary>
        /// Annotation info added by analyzers
        /// </summary>
        public IList<AnnotationInfoItem> AnnotationInfo { get; set; }

        /// <summary>
        /// Custom attributes based on needs of the dependency registry
        /// </summary>
        public IDictionary<string, object> CustomAttributes { get; set; }

        /// <summary>
        /// List of sub-dependecies for this dependency
        /// Represented using wrapper for the edge
        /// </summary>
        public IList<DependencyEdge> Dependencies { get; set; }

        /// <summary>
        /// Compares two dependency items alphabeticaly using id and version
        /// </summary>
        /// <param name="other">Other dependency toeb compared with</param>
        /// <returns></returns>
        public int CompareTo(DependencyItem other)
        {
            string key1 = $"{this.Id}{this.Version}";
            string key2 = $"{other.Id}{other.Version}";
            return key1.CompareTo(key2);
        }

        /// <summary>
        /// Prints the dependency tree in console
        /// </summary>
        /// <param name="indent">Indentation of the node</param>
        /// <param name="last">Marks the item as last, prints end of this branch</param>
        /// <param name="edgeInfo">Additional information about the edge</param>
        public string PrintPretty(string indent, bool last, string edgeInfo = "") 
        {
            string annotationindent = indent;
            string dataResult = "";
            //Console.Write(indent);
            dataResult += indent;
            if (last)
            {
                //Console.Write("\\-");
                dataResult += "\\-";
                indent += "  ";
            }
            else
            {
                //Console.Write("|-");
                dataResult += "|-";
                indent += "| ";
            }
            //Console.WriteLine(this);
            dataResult += this.ToString();              //Prints vertex
            if(edgeInfo != null && edgeInfo != "")
            {
                dataResult += $" ({edgeInfo})";         //Adds edge info
            }
            dataResult += "\n";                         //Newline

            foreach (var annotationData in AnnotationInfo)
            {
                dataResult += indent + "|---:" + annotationData + "\n";
            }

            for (int i = 0; i < Dependencies.Count; i++)
            {
                dataResult += Dependencies[i].Link.PrintPretty(indent, i == Dependencies.Count - 1, Dependencies[i].RequiredVersionsText);
            }

            return dataResult;
        }


        /// <summary>
        /// Prints object in string format
        /// </summary>
        /// <returns>Textual representation of the DependencyItem</returns>
        public override string ToString()
        {
            return $"Id: {Id}, Name: {Name}, Version: {Version}, Direct dependencies: {Dependencies.Count}";
        }
    }
}
