﻿using CommandLine;
using Janus.Core.Models;
using Janus.Core.Utils;
using Janus.Parsers.DotnetCoreNuget.Parser;
using Newtonsoft.Json;
using NuGet.Packaging;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading.Tasks;

namespace Janus.Parsers.DotnetCoreNuget
{
    class Program
    {
        /// <summary>
        /// Command line arguments parsed as options
        /// </summary>
        public static CLIOptions options;

        /// <summary>
        /// Instance of the nuget wrapper
        /// </summary>
        private static NugetAPI.NugetWrapper _nugetWrapper = NugetAPI.NugetWrapper.Instance;

        /// <summary>
        /// Cache of the visited dependencies to be linked. This cache reduces the API calls to absolute minimum
        /// Key:   Tuple<IdPackage, VersionPackage>
        /// Value: objectReference
        /// </summary>
        private static Dictionary<Tuple<string, string>, DependencyItem> _cachedDependencies = new Dictionary<Tuple<string, string>, DependencyItem>();


        /// <summary>
        /// Main entry point of the program
        /// </summary>
        /// <param name="args">CLI Arguments</param>
        /// <returns></returns>
        static async Task Main(string[] args)
        {
            //Console.WriteLine("Started parser, DotnetCoreNuget!");
            //NugetAPI.NugetWrapper nugetWrapper = NugetAPI.NugetWrapper.Instance;
            //await nugetWrapper.GetMetadataAsync("Newtonsoft.Json", "13.0.1");
            //await nugetWrapper.GetDependencyDataAsync("Newtonsoft.Json", "13.0.1", "net5.0");
            //await nugetWrapper.GetDependencyDataAsync("QuikGraph.Graphviz", "2.3.1", "net5.0");

            var result = await CommandLine.Parser.Default.ParseArguments<CLIOptions>(args)
                                                            .WithParsedAsync(RunOptions);
            result.WithNotParsed(HandleParseError);
        }

        /// <summary>
        /// Main logic, if argument parsing is without an error
        /// </summary>
        /// <param name="opts">CLI Arguments as options</param>
        /// <returns></returns>
        static async Task RunOptions(CLIOptions opts)
        {
            options = opts; //Pass args to the rest of the program

            IList<SolutionItem> solutions = new List<SolutionItem>();
            IList<ProjectItem> projects = new List<ProjectItem>();
            IList<DependencyItem> dependencies = new List<DependencyItem>();

            if (opts.InputDependencies.Any())
            {
                if (string.IsNullOrEmpty(opts.FrameworkVersion))
                {
                    //FW Version not specified!!!
                    Console.Error.WriteLine($"Input dependencies specified, but framework this option requires framework version to be also specified!");
                    Console.Error.WriteLine("Skipping...");
                    return;
                }
                foreach (var inputDep in options.InputDependencies)
                {
                    DependencyItem item = await ProcessInputDependency(inputDep, options.SearchDepth);
                    if(item != null)
                    {
                        dependencies.Add(item);
                    }
                }
            }
            else if (opts.InputFiles.Any())
            {
                foreach (var inputFile in opts.InputFiles)
                {
                    if (!File.Exists(inputFile))
                    {
                        Console.Error.WriteLine($"File {inputFile} does not exist!");
                        Console.Error.WriteLine("Skipping...");
                        continue;
                    }

                    //File exists ... we can process to parsing

                    switch (Path.GetExtension(inputFile))
                    {
                        case ".csproj":
                            ProjectItem projItem = await ProcessProject(inputFile);
                            projects.Add(projItem);
                            break;
                        case ".sln":
                            SolutionItem solutionItem = await ProcessSolution(inputFile);
                            solutions.Add(solutionItem);
                            break;
                        default:
                            Console.Error.WriteLine($"File {inputFile} has unknown extension!");
                            Console.Error.WriteLine("Supported extensions are \".csproj\" and \".sln\"!");
                            Console.Error.WriteLine("Skipping...");
                            break;
                    }
                }

                if (solutions.Count > 0 && projects.Count > 0)
                {
                    //TODO: Maybe can be solved, somehow?
                    //Combined .. why? NOT SUPPORTED!
                    Console.Error.WriteLine($"Mixed input detected, unable to serialize!");
                    Console.Error.WriteLine("You can not combine the solution and project in the same job!");
                    Console.Error.WriteLine("Skipping printing...");
                    return;
                }
            }
            else
            {
                //No valueable input
                //var arguments = CommandLine.Parser.Default.FormatCommandLine(options);
                Console.Error.WriteLine($"Nothing to work with!");
                Console.Error.WriteLine($"Please specify the input using command line arguments!");
                Console.Error.WriteLine($"For more information use --help");
            }

            if (solutions.Count > 0)
            {
                //Print solutions
                DataWorker.Print(solutions, options);
            }
            else if(projects.Count > 0)
            {
                //Print projects
                DataWorker.Print(projects, options);
            }
            else if (dependencies.Count > 0)
            {
                //Print dependencies
                DataWorker.Print(dependencies, options);
            }
            else
            {
                Console.Error.WriteLine($"Nothing to print!");
                Console.Error.WriteLine("Skipping printings...");
            }
        }

        /// <summary>
        /// This method is called in case of argument parsing failure
        /// </summary>
        /// <param name="errs">Parsing errors as text</param>
        static void HandleParseError(IEnumerable<Error> errs)
        {
            Console.Error.WriteLine("ERROR During arguments matching ...");
            foreach (var item in errs)
            {
                Console.Error.WriteLine(item);
            }
        }

        /// <summary>
        /// Processes the dotnetcore solution for the depedendecy analysis, but first parses the projects in the solution
        /// </summary>
        /// <param name="solutionPath">Path to the solution file</param>
        /// <returns>Solution structure with parsed information</returns>
        static async Task<SolutionItem> ProcessSolution(string solutionPath)
        {
            //Parse file with direct deps
            SolutionItem solution = DotnetCoreParser.ParseSolution(solutionPath);

            await CrawlSubDependenciesForSolution(solution, options.SearchDepth);

            return solution;
        }

        /// <summary>
        /// Processes the dotnetcore project for the depedendecy analysis
        /// </summary>
        /// <param name="projectPath">Path to the project file</param>
        /// <returns>Project structure with parsed information</returns>
        static async Task<ProjectItem> ProcessProject(string projectPath)
        {
            //Parse file with direct deps
            ProjectItem project = DotnetCoreParser.ParseProject(projectPath);

            await CrawlSubDependeciesForProject(project, options.SearchDepth);

            return project;
        }

        /// <summary>
        /// Processes one dependency defined using the input argument
        /// Input format: "<id>;<version>"
        /// </summary>
        /// <param name="dependencyInfo">Info about the dependency, id and version</param>
        /// <param name="depth">Dept of the dependency tree, special value -1 means maximum</param>
        /// <returns>Dpeendency item filled with the data, in case of problem, null</returns>
        static async Task<DependencyItem> ProcessInputDependency(string dependencyInfo, int depth = -1)
        {
            //Parse dep info
            // Format "<id>;<version>"
            var splitText = dependencyInfo.Split(";");
            if(splitText.Length != 2)
            {
                //Error ... wrong format
                Console.Error.WriteLine($"Error: Wrong dependency info format provided, problem is with the: \"{dependencyInfo}\" (Skipping ...)");
                return null;
            }
            DependencyItem depItem = await _nugetWrapper.FillMetadataAsync(splitText[0], splitText[1], options.RepositoryURL); //TODO: Fix this ugly link to the repo url

            await CrawlSubDependeciesForDependency(depItem, options.FrameworkVersion, depth);

            return depItem;
        }

        /// <summary>
        /// Crawl subdependencies, with optional depth limit
        /// Contacts Nuget server using wrapper and obtains dependencies
        /// </summary>
        /// <param name="solution">Solution to be dependencies crawled for</param>
        /// <param name="depth">Dept of the dependency tree, special value -1 means maximum</param>
        /// <returns></returns>
        static async Task CrawlSubDependenciesForSolution(SolutionItem solution, int depth = -1)
        {
            foreach (var project in solution.Projects)
            {
                if(project.Link != null)
                {
                    await CrawlSubDependeciesForProject(project.Link, depth);
                }
            }
        }

        /// <summary>
        /// Crawl subdependencies, with optional depth limit
        /// Contacts Nuget server using wrapper and obtains dependencies
        /// </summary>
        /// <param name="project">Project to be dependencies crawled for</param>
        /// <param name="depth">Dept of the dependency tree, special value -1 means maximum</param>
        /// <returns></returns>
        static async Task CrawlSubDependeciesForProject(ProjectItem project, int depth = -1)
        {
            if(depth == 1)
            {
                //Only direct, no need to analyze deeper
                return;
            }

            foreach (var dependencyEdge in project.Dependencies)
            {
                if(dependencyEdge.Link != null)
                {
                    await CrawlSubDependeciesForDependency(dependencyEdge.Link, project.LanguageVersion, depth);
                }
            }
        }

        /// <summary>
        /// Crawl subdependencies, with optional depth limit
        /// Contacts Nuget server using wrapper and obtains dependencies
        /// </summary>
        /// <param name="dependencyToBeCrawled">Dependency to be sub-dependencies crawled for</param>
        /// <param name="depth">Dept of the dependency tree, special value -1 means maximum</param>
        /// <returns></returns>
        static async Task CrawlSubDependeciesForDependency(DependencyItem dependencyToBeCrawled, string frameworkVersion, int depth = -1)
        {
            Queue<DependencyItem> toBeCrawled = new Queue<DependencyItem>();

            int currentDepth = 0;

            toBeCrawled.Enqueue(dependencyToBeCrawled); //First item to be processed

            while (toBeCrawled.Any() && (currentDepth < depth || depth == -1))
            {
                DependencyItem current = toBeCrawled.Dequeue();

                var index = new Tuple<string, string>(current.Id, current.Version);

                //First check if the dependency is not in cache
                if (_cachedDependencies.ContainsKey(index))
                {
                    //Already in cache!!! Just reference that and continue
                    //Can result in depth problem???
                    dependencyToBeCrawled.Dependencies = _cachedDependencies[index].Dependencies;
                    continue;
                }

                _cachedDependencies[index] = current;
                List<DependencyEdge> crawledItems = await GetDependencies(current, frameworkVersion, _nugetWrapper);

                foreach (var oneCrawledItemEdge in crawledItems)
                {
                    var oneCrawledItem = oneCrawledItemEdge.Link;

                    var crawledItemIndex = new Tuple<string, string>(oneCrawledItem.Id, oneCrawledItem.Version);
                    if (!_cachedDependencies.ContainsKey(crawledItemIndex))
                    {
                        //New dependency, has to be resolved?
                        var toBeResolvedMatch = toBeCrawled.Where(o => (o.Id == oneCrawledItem.Id && o.Version == oneCrawledItem.Version));
                        if (toBeResolvedMatch.Any())
                        {
                            //Is included in to do list ... just reference
                            oneCrawledItemEdge.Link = toBeResolvedMatch.First();
                            current.Dependencies.Add(oneCrawledItemEdge);   //Add to the deps list the existing reference
                        }
                        else
                        {
                            //Completely new ... will be included in queue
                            current.Dependencies.Add(oneCrawledItemEdge);   //Add to the deps list
                            toBeCrawled.Enqueue(oneCrawledItem);            //Add to the to resolve queue

                        }
                    }
                }
                currentDepth++;
            }

        }

        /// <summary>
        /// Obtains dependencies for the specific dependency item. With specific framework and nugetWrapper instance
        /// </summary>
        /// <param name="dependency">Dependency from which the subdependencies should be obtained</param>
        /// <param name="frameworkVersion">Version of the framework for exact match</param>
        /// <param name="nugetWrapper">Nuget wrapper instance for calls</param>
        /// <returns></returns>
        static async Task<List<DependencyEdge>> GetDependencies(DependencyItem dependency, string frameworkVersion, NugetAPI.NugetWrapper nugetWrapper)
        {
            List<DependencyEdge> results = new List<DependencyEdge>();

            var packageId = dependency.Id;
            var version = dependency.Version;
            var framework = frameworkVersion;

            var depInfo = await nugetWrapper.GetDependencyDataAsync(packageId, version, framework, options.RepositoryURL); //TODO: Fix this ugly link to the repo url

            foreach (var item in depInfo.Dependencies)
            {
                DependencyItem depItem = null;
                var index = new Tuple<string, string>(item.Id, item.VersionRange.MinVersion.ToString());

                //First check if the dependency is not in cache
                if (_cachedDependencies.ContainsKey(index))
                {
                    //Already in cache!!! Just reference that and continue
                    depItem = _cachedDependencies[index];
                }
                else
                {
                    depItem = await nugetWrapper.FillMetadataAsync(item.Id, item.VersionRange.MinVersion.ToString(), options.RepositoryURL); //TODO: Fix this ugly link to the repo url
                }
                 
                //Create edge
                DependencyEdge edgeItem = new DependencyEdge();
                edgeItem.Parent = dependency;
                edgeItem.Link = depItem;
                if (item.VersionRange.HasLowerBound)
                {
                    if (item.VersionRange.IsMinInclusive)
                    {
                        edgeItem.VersionLimits.Add(new VersionLimit() { RequiredVersion = item.VersionRange.MinVersion.ToString(), LimitSymbol = VersionLimit.OperatorType.GreaterOrEqual});
                    }
                    else
                    {
                        edgeItem.VersionLimits.Add(new VersionLimit() { RequiredVersion = item.VersionRange.MinVersion.ToString(), LimitSymbol = VersionLimit.OperatorType.Greater });
                    }
                }
                if (item.VersionRange.HasUpperBound)
                {
                    if (item.VersionRange.IsMaxInclusive)
                    {
                        edgeItem.VersionLimits.Add(new VersionLimit() { RequiredVersion = item.VersionRange.MaxVersion.ToString(), LimitSymbol = VersionLimit.OperatorType.LowerOrEqual });

                    }
                    else
                    {
                        edgeItem.VersionLimits.Add(new VersionLimit() { RequiredVersion = item.VersionRange.MaxVersion.ToString(), LimitSymbol = VersionLimit.OperatorType.Lower });
                    }
                }

                results.Add(edgeItem);
            }

            return results;
        }
    }
}
